#!/bin/perl
#
# cbr-replictor.pl
#
# Programmed by G. Polhill 13 October 2003
# Designed by L. Izquierdo 10 October 2003
#
# This is a script to repeat and analyse results from a model of social games
# using agents with case-based reasoning cognitive architecture.
#
# Usage:
#
# ./cbr-replicator.pl <mode> [<runs> | <filename>] <prog> <param> 
#                         [<cmd param...>]
#
# where <mode> is either 'Deterministic' or 'Random'. If 'Random' then number
# of runs or a file name with seeds must be specified.
#
# <prog> is the full path to the model to run.
#
# <param> is the parameter file to use
#
# <cmd param> is an optional list of command line parameters to give to the
# model. Command line parameters beginning with + should be specified first,
# and those with - last.

use strict;

# Set up a log file to record usage

my ($s, $m, $h, $d, $mo, $y) = gmtime;
$mo++;
$y += 1900;

my $logfilename = sprintf("cbr-rep-%04d%02d%02d-%02d%02d%02d.log", $y, $mo, $d,
			  $h, $m, $s);

open(LOG, ">$logfilename") or die "Cannot create log file $logfilename: $!\n";

my $logfh = select(LOG);
$| = 1;				# Make LOG autoflush
select($logfh);

&log("Command line: $0 ".join(" ", @ARGV));
&log("Process ID: $$");
&log("User effective (real): ".(getpwuid($>))[0]." (".(getpwuid($<))[0].")");
&log("Perl location (version, platform): $^X ($^V, $^O)");
&log("Platform: ".`uname -a`);

# Parse the command line options

&usage if($#ARGV < 2);		# Check the minimum number of options are there

my $mode = shift(@ARGV);	# Get the mode
$mode =~ tr/[A-Z]/[a-z]/;	# Shift it to lowercase

my $runs;
my $seedfile;
my @seeds;
if($mode eq 'random') {
  $runs = shift(@ARGV);
  &usage if($#ARGV < 1);
  if($runs =~ /^\d+$/) {
    $seedfile = $logfilename;
    $seedfile =~ s/\.log$/.seeds/;
    open(SEEDS, ">$seedfile") or die "Cannot create seed file $seedfile: $!\n";
    my $seedfh = select(SEEDS);
    $| = 1;			# Make SEEDS autoflush
    select($seedfh);
    $mode .= '.swarmseeds';
  }
  else {
    $seedfile = $runs;
    open(SEEDS, "<$seedfile") or die "Cannot read seed file $seedfile: $!\n";
    $runs = 0;
    while(my $line = <SEEDS>) {
      chomp $line;
      if($line =~ /^(\d+)$/) {
	push(@seeds, $1);
	$runs++;
      }
    }
    close(SEEDS);
    $mode .= '.fileseeds';
  }
}
elsif($mode ne 'deterministic') {
  &usage;
}

my $prog = shift(@ARGV);
my $param = shift(@ARGV);
my @modelparam;
my @swarmparam;
&get_model_swarm_param(\@modelparam, \@swarmparam, @ARGV);
my $cdstrlen;
$cdstrlen = &get_cd_string_length($param) if($mode eq 'deterministic');

if($mode =~ /^random/) {
  &log("Mode is RANDOM: $runs runs");
  if($mode =~ /fileseeds$/) {
    &log("Seeds for runs are loaded from file $seedfile");
  }
  else {
    &log("Seeds for runs are generated by Swarm, and will be saved in file "
	 ."$seedfile");
  }
}
else {
  &log("Mode is DETERMINISTIC: $cdstrlen C/D string length");
}
&log("Model command-line parameters: ".join(" ", @modelparam));
&log("Swarm command-line parameters: ".join(" ", @swarmparam));


# Main loop -- this is where the models get run and the results analysed

my $ok = 1;
my @cdarg;
if($mode eq 'deterministic') {
  for(my $i = 0; $i < $cdstrlen; $i++) {
    $cdarg[$i] = 'C';
  }
}
elsif($runs == 0) {
  $ok = 0;
}

my $run = 0;
my %rewards;
my $stableCooperations = 0;
my $stableDefections = 0;
my @payoffs;

while($ok) {
  # Set up the command and create the output file name
  my $cmd = "$prog ";
  $cmd .= join(" ", @modelparam);
  if($mode eq 'deterministic') {
    $cmd .= " +d ".join("", @cdarg);
  }
  $cmd .= " $param ";
  $cmd .= join(" ", @swarmparam);
  my $outfile;
  if($mode eq 'random.swarmseeds') {
    $cmd .= " -s";
    $outfile = sprintf("run-%05d.txt", $run + 1);
  }
  elsif($mode eq 'random.fileseeds') {
    $cmd .= " -S $seeds[$run]";
    $outfile = "$seeds[$run].txt";
  }
  else {
    $outfile = join("", @cdarg).".txt";
  }
  &log("Running command $cmd");
  &log("Output file $outfile");
  open(OUT, ">$outfile") or die "Cannot create output file $outfile: $!\n";

  # Run the command and process its output
  open(CMD, "$cmd |") or die "Cannot run command $cmd: $!\n";
  while(my $line = <CMD>) {
    chomp $line;
    if($mode eq 'random.swarmseeds') {
      if($line =~ /^Seed: (\d+)$/) {
	print SEEDS "$1\n";
      }
    }
    if($line =~ /Reward given (\d+) out of (\d+) times/) {
      my ($num, $den) = ($1, $2);
      if(defined($rewards{$num}{$den})) {
	$rewards{$num}{$den}++;
      }
      else {
	$rewards{$num}{$den} = 1;
      }
      &log("Reward $num/$den is now $rewards{$num}{$den}");
    }
    if($line =~ /STABLE COOPERATION/) {
	$stableCooperations++;
    }
    if($line =~ /STABLE DEFECTION/) {
	$stableDefections++;
    }
    if($line =~ /Average Group Payoff: (\S+)$/) {
	push(@payoffs, $1);
    }
    print OUT "$line\n";
  }
  close(CMD);
  close(OUT);
  &log("Finished command $cmd");
  $run++;
 
  # Iterate
  if($mode eq 'deterministic') {
    $ok = 0;
    for(my $i = 0; $i <= $#cdarg; $i++) {
      if($cdarg[$i] eq 'C') {
	$cdarg[$i] = 'D';
	$ok = 1;
	last;
      }
      else {
	$cdarg[$i] = 'C';
      }
    }
  }
  else {			# Random mode
    $ok = ($run eq $runs) ? 0 : 1;
  }
}

if($mode eq 'random.swarmseeds') {
  close(SEEDS);
}

# Create a results file and open it

my $resultsfile = $logfilename;
$resultsfile =~ s/\.log/-results.txt/;
open(RESULTS, ">$resultsfile")
  or die "Error creating results file $resultsfile: $!\n";

print RESULTS "Stable Cooperations: \t$stableCooperations times\n";
print RESULTS "Stable Defections: \t$stableDefections times\n";

# Calculate the reward rate

my $reward_rate = 0;
my $beta_sum = 0;
while(my($num, $value) = each(%rewards)) {
  while(my($den, $beta) = each(%$value)) {
    $reward_rate += ($num / $den) * $beta;
    $beta_sum += $beta;
    print RESULTS "Reward given \t$num\tout of \t$den\t times:\t$beta\n";
  }
}
&log("$run runs completed");
$reward_rate /=  $beta_sum;
print RESULTS "\nTotal:\t$beta_sum\nAverage reward rate:\t$reward_rate\n";

#Calculate the standard error of the average reward rate
my $base;
my $square;
my $element;
my $stdError = 0;

while(my($num, $value) = each(%rewards)) {
  while(my($den, $beta) = each(%$value)) {
    $base = ($num / $den) - $reward_rate;
    $square = $base**2;
    $element = $square * $beta;
    $stdError += $element;
  }
}

#Variance of the variable: reward rate
$stdError /= ($beta_sum -1);

#Variance of the variable: average reward rate
$stdError /= $beta_sum;

#Standard error of the variable: average reward rate
$stdError = sqrt($stdError);

print RESULTS "Standard Error of the average reward rate:\t$stdError\n";

#Calculate the average group payoff
my $averagePayoff = 0;
my $numberOfPayoffs = scalar(@payoffs);

for(my $i = 0; $i <= $#payoffs; $i++) {
  $averagePayoff += $payoffs[$i];
}
$averagePayoff /= $numberOfPayoffs;

print RESULTS "\nTotal:\t$numberOfPayoffs\n";
print RESULTS "Average Group Payoff:\t$averagePayoff\n";

#and the standard error of the average group payoff
my $stdErrorPayoff = 0;
for(my $i = 0; $i <= $#payoffs; $i++) {
  $stdErrorPayoff += ($payoffs[$i] - $averagePayoff) ** 2;
}

#Variance of the variable: group payoff
$stdErrorPayoff /= $numberOfPayoffs - 1;

#Variance of the variable: average group payoff
$stdErrorPayoff /= $numberOfPayoffs;

#Standard error of the variable: average group payoff
$stdErrorPayoff = sqrt($stdErrorPayoff);

print RESULTS "Standard Error of the average Group Payoff:\t";
print RESULTS "$stdErrorPayoff\n";

close(RESULTS);
&log("Results saved in $resultsfile");

# Clean up and exit
  
close(LOG);
exit 0;

###############################################################################

# log
#
# Put an entry in the log file

sub log {
  my ($entry) = @_;

  my ($s, $m, $h, $d, $mo, $y) = localtime;
  $mo++;
  $y += 1900;

  print LOG sprintf("%04d%02d%02d-%02d%02d%02d $entry\n",
		    $y, $mo, $d, $h, $m, $s);
}

# usage
#
# Print a usage message to STDERR and then exit

sub usage {
  print STDERR <<USAGE;

This is a program to make multiple runs of a model of social games using agents
with case-based reasoning for their cognitive architecture and then analyse the
output of those runs.

Usage:
\t$0 <mode> [<runs> | <filename>] <prog> <param> [<cmd params...>]

\t\t<mode>         'Random' or 'Deterministic' 
\t\t               (use Deterministic only if forwardMemory is 1).
\t\t<runs>         If mode is 'Random' specify the number of runs
\t\t<filename>     If mode is 'Random' specify a file with seeds
\t\t<prog>         Full path to model program to use
\t\t<param>        Parameter file to use with the model
\t\t<cmd param...> Command line parameters to send to the program
\t\t\tThese should be with all + (model specific) options
\t\t\tfirst then - (Swarm specific) options last.
USAGE
  &log("Error!");
  close(LOG);
  exit 1;
}

# get_model_swarm_param(model, swarm, args)
#
# Separate the model specific command line options and the swarm specific
# command line options from the list of arguments supplied.

sub get_model_swarm_param {
  my ($model, $swarm, @args) = @_;

  my $cmdmode = 'model';
  my $firstswarmarg = 'undef';

  foreach my $arg (@args) {
    if($cmdmode eq 'swarm') {
      if($arg =~ /^\+/) {
	print STDERR "Model command line argument $arg after Swarm argument "
	  ."begins ($firstswarmarg)\n";
	&usage;
      }
    }
    elsif($arg =~ /^-/) {
      $firstswarmarg = $arg;
      $cmdmode = 'swarm';
    }

    if($cmdmode eq 'model') {
      push(@$model, $arg);
    }
    else {
      if($arg eq '-S' || $arg eq '-s') {
	print STDERR "You cannot pass seed arguments to Swarm from the "
	  ."command line to this program, as it will take care of this.\n";
	&usage;
      }
      push(@$swarm, $arg);
    }
  }
}

# get_cd_string_length(paramfile)
#
# Read the parameter file to get the parameters determining the string length
# then calculate the string length and return it.

sub get_cd_string_length {
  my ($paramfile) = @_;

  open(PARAM, "<$paramfile") or die "Cannot open $paramfile for reading: $!\n";

  &log("Reading parameters from file $paramfile");
  my %var;
  while(my $line = <PARAM>) {
    chomp $line;
    if($line =~ /^\s+\#:(\S+)\s+(\S+)$/) {
      $var{$1} = $2;
      &log("Read parameter $1 = $2");
    }
  }
  close(PARAM);

  my $error = 0;
  foreach my $needed_var ('xSize', 'ySize', 'descriptorOtherDefectorsStr',
			  'descriptorMyDecisionsStr', 'bMemory', 'fMemory') {
    if(!defined($var{$needed_var})) {
      print STDERR ", " if $error;
      print STDERR $needed_var;
      $error = 1;
    }
  }

  if($error) {
    print STDERR " not found in parameter file $paramfile\n";
    usage;
  }

  my $nAgentsPow;
  my $twoPow;
  if($var{'descriptorOtherDefectorsStr'} eq 'NO') {
    $nAgentsPow = 0;
  }
  elsif($var{'descriptorOtherDefectorsStr'} eq 'YES') {
    $nAgentsPow = $var{'bMemory'};
  }
  else {
    print STDERR "Invalid setting for descriptorOtherDefectorsStr parameter:"
      ." $var{'descriptorOtherDefectorsStr'}\n";
    &usage;
  }

  if($var{'descriptorMyDecisionsStr'} eq 'NO') {
    $twoPow = 0;
  }
  elsif($var{'descriptorMyDecisionsStr'} eq 'YES') {
    $twoPow = $var{'bMemory'};
  }
  else {
    print STDERR "Invalid setting for descriptorMyDecisionsStr parameter:"
      ." $var{'descriptorMyDecisionsStr'}\n";
    &usage;
  }

  my $nAgents = $var{'xSize'} * $var{'ySize'};
  return ($nAgents * ($var{'bMemory'}
		      + (2 ** $twoPow) * ($nAgents ** $nAgentsPow)));
}

